"""
Copyright (c) 2023, Magentix
This code is licensed under simplified BSD license (see LICENSE for details)
"""
import json
import os
import re
import time
import unicodedata
from html import escape, unescape

VERSION = '1.2.3'


def custom_http_response(response: tuple or None, args: dict) -> tuple or None:
    """
    Test if the requested path matches with an editor page
    """
    editor = Editor(
        args['path'],
        args['request'],
        args['_stapy']['file_system'],
        args['_stapy']['json_query'],
        args['_stapy']['generator'],
        args['_stapy']['plugins']
    )
    return editor.match() or response


class Editor:
    _config = {}
    _messages = {}
    _editor_key = None

    def __init__(self, path, request, file_system, json_query, generator, plugins) -> None:
        self._path = path or '/'
        self._request = request
        self._fs = file_system
        self._jq = json_query
        self._gs = generator
        self._sp = plugins
        self._load_config(path.split('/')[1])

    def match(self) -> tuple or None:
        """
        Test if the requested path matches with an editor page
        """
        if not self._editor_key:
            return None

        if self._path.rstrip('/') == self._editor_key:
            return self.page_list(), 'text/html'

        if self._path.rstrip('/') == f'{self._editor_key}/build':
            self._build()
            return self.page_list(), 'text/html'

        if self._path.rstrip('/') == f'{self._editor_key}/add':
            return self.page_edit(), 'text/html'

        remove = re.search(rf'^{self._editor_key}/remove/(.*?)$', self._path, flags=re.IGNORECASE)
        if remove:
            self._page_remove(remove.group(1))
            return self.page_list(), 'text/html'

        edit = re.search(rf'^{self._editor_key}/edit/(.*?)$', self._path, flags=re.IGNORECASE)
        if edit:
            path = edit.group(1)
            if self._request.get('form_method', [''])[0] == 'save':
                path = self._page_save(path)
                if self._request.get('ajax', ['0'])[0] == '1':
                    return self._block_messages(), 'text/html'
            return self.page_edit(path or None), 'text/html'

        if self._path.rstrip('/') == f'{self._editor_key}/style.css':
            file = f'{self._fs.get_plugins_dir()}{os.sep}editor{os.sep}style.css'
            with open(os.path.normpath(file), 'r', encoding='utf-8') as css:
                return css.read(), 'text/css'

        if self._path.rstrip('/') == f'{self._editor_key}/app.js':
            file = f'{self._fs.get_plugins_dir()}{os.sep}editor{os.sep}app.js'
            with open(os.path.normpath(file), 'r', encoding='utf-8') as script:
                return script.read(), 'application/javascript'

        return self._sp.dispatch('editor_plugin_match', None, False, {'editor': self})

    def add_message(self, level: str, message: str):
        """
        Add a message
        """
        self._messages[level] = [message]

    def get_messages(self, level: str, reset: bool = True):
        """
        Retrieve messages by level
        """
        messages = dict(self._messages)
        if reset:
            self._messages[level] = []
        return messages.get(level, [])

    def get_config(self) -> dict:
        """
        Retrieve configuration
        """
        return self._config

    def get_path(self) -> str:
        """
        Retrieve path
        """
        return self._path

    def get_editor_key(self) -> str or None:
        """
        Retrieve editor key
        """
        return self._editor_key

    def _load_config(self, key) -> dict:
        """
        Load the editor config file
        """
        file = self._fs.get_source_dir() + os.sep + 'editor.json'
        self._config = {}
        if os.path.exists(file):
            config = self._fs.merge_json([file]).get(key, {})
            if config:
                self._editor_key = '/' + key
            self._config = config
        return self._config

    def _build(self):
        """
        Build the website for all environments
        """
        result = self._gs.build()
        if not result:
            self.add_message('error', 'Nothing to build. You need to add an environment in the web directory.')
        else:
            messages = []
            for env, data in result.items():
                messages.append(
                    '[' + env + '] ' + str(data['number']) + ' pages generated in ' + str(data['time']) + ' seconds'
                )
            self.add_message('success', '<br />'.join(messages))

    def _page_remove(self, current_path) -> str:
        """
        Remove the page with all child blocks
        """
        if not current_path:
            self.add_message('error', 'The page path is missing')
            return current_path

        if not self._can_edit_page(current_path):
            self.add_message('error', 'The page cannot be deleted')
            return current_path

        json_path = self._fs.get_page_config(current_path)
        if not os.path.exists(json_path):
            self.add_message('error', f'Page "{current_path}" does not exist')
            return current_path

        page = self._fs.get_page_data(current_path, False)
        config = self.get_config().get('form', {})

        if not config.get('actions', {}).get('delete', False):
            self.add_message('error', 'Access denied')
            return current_path

        # Remove the page child blocks
        for fieldset in config.get('fieldsets', []):
            for var, options in fieldset.get('fields', {}).items():
                if var in page and options.get('type') == 'block':
                    file = self._fs.get_source_dir(page[var])
                    if os.path.exists(file):
                        os.remove(file)

        # Remove the page config file
        os.remove(json_path)

        self.add_message('success', f'Page "{current_path}" has been deleted')

        return current_path

    def _page_save(self, current_path) -> str:
        """
        Save the posted data in the page JSON data for the path (pages/{path}.json
        """
        if '_full_path' not in self._request:
            self.add_message('error', 'Page path is missing')
            return current_path
        if '_page_type' not in self._request:
            self.add_message('error', 'Page type is missing')
            return current_path

        config = self.get_config().get('form', {})

        if not config.get('actions', {}).get('save', False):
            self.add_message('error', 'Access denied')
            return current_path

        # Clean the page path
        page_type = self._request.get('_page_type')[0]
        page_extension = '.' + ('html' if page_type in ['md'] else page_type)
        new_path = self._format_path((self._request.get('_full_path')[0]).replace(page_extension, ''))
        if not new_path:
            self.add_message('error', 'Page path is empty')
            return current_path
        new_path += page_extension

        if not self._can_edit_page(new_path):
            self.add_message('error', 'The page cannot be edited')
            return current_path

        new_full_path = self._fs.get_page_config(new_path)
        current_full_path = self._fs.get_page_config(current_path)

        # Test if the new page path already exists
        if not current_path and os.path.exists(new_full_path):
            self.add_message('error', 'Page path already exists')
            return current_path

        # Test if the updated page path already exists
        if current_path and new_full_path != current_full_path and os.path.exists(new_full_path):
            self.add_message('error', 'Page path already exists')
            return current_path

        # Load JSON page data
        data = self._fs.merge_json([current_full_path])\
            if current_path and os.path.exists(current_full_path)\
            else json.loads('{}')

        # Loop in config fields and add it to JSON data if required
        for fieldset in config.get('fieldsets', []):
            for var, options in fieldset.get('fields', {}).items():
                if not options.get('save', True) or var not in self._request:
                    continue
                values = self._request[var]
                if options.get('type') in ['file', 'image']:
                    if var + '_remove' in self._request and self._request[var + '_remove'][0] == '1':
                        if data.get(var, None) and os.path.exists(self._fs.get_source_dir('assets/' + data[var])):
                            os.unlink(self._fs.get_source_dir('assets/' + data[var]))
                        data[var] = ''
                    path = self._save_file(values[0], options)
                    if path:
                        if data.get(var, None)\
                                and path != data[var]\
                                and os.path.exists(self._fs.get_source_dir('assets/' + data[var])):
                            os.unlink(self._fs.get_source_dir('assets/' + data[var]))
                        data[var] = path
                elif options.get('type') == 'multiselect':
                    data[var] = list(filter(None, values))
                elif options.get('type') == 'block':
                    data[var] = values[0].replace('\r\n', '\n')
                    extension = options.get('mode', 'html')
                    if extension[0] == '$':
                        extension = self._request.get(extension.lstrip('$'), ['html'])[0]
                    if extension == 'html':
                        code_elements = self.__extract_code(data[var])
                        for code in code_elements:
                            data[var] = data[var].replace(
                                '<code' + code.get('attr') + '>' + code.get('content') + '</code>',
                                '<code' + code.get('attr') + '>' + escape(code.get('content')) + '</code>'
                            )
                    save_dir = options.get('save_dir', '').strip('/')
                    content_file = os.path.splitext(new_path)[0] + '.' + extension
                    self._fs.create_file(self._fs.get_source_dir(save_dir) + os.sep + content_file, data[var])
                    data[var] = (save_dir + '/' if save_dir else '') + content_file
                else:
                    data[var] = values[0] if options.get('allow_html', False) else escape(values[0], False)

        # Remove the old page path if path has been updated
        if current_path and new_full_path != current_full_path:
            self._page_remove(current_path)

        # Save the JSON data
        self._fs.create_file(new_full_path, json.dumps(data, indent=2, ensure_ascii=False))
        self.add_message('success', 'The page has been saved')

        return new_path

    def _save_file(self, upload: dict, options: dict) -> str:
        """
        Save the file and return the path from the asset directory
        """
        if isinstance(upload, dict) and upload.get('filename', ''):
            file = self._format_path(upload.get('filename'), '[^0-9a-z.]')
            extension = self._fs.get_file_extension(file)
            if not options.get('extensions') or extension in options.get('extensions'):
                path = (options.get('directory', '').strip('/') + '/' + str(int(time.time())) + '-' + file).strip('/')
                self._fs.create_directory(self._fs.get_source_dir('assets/' + path))
                with open(self._fs.get_source_dir('assets/' + path), 'wb') as file:
                    file.write(upload.get('content'))
                return path
            self.add_message('error', options.get('label') + ': "' + extension + '" is not allowed')
        return ''

    def page_list(self) -> str:
        """
        List all the pages in a table
        """
        config = self.get_config().get('grid', {})

        # Load the pages
        pages = self._jq.fetch(config.get('query', ''), True)

        # Build the table head
        table_head = '<tr><th>Path</th>'
        for key, options in config.get('columns', {}).items():
            table_head += f'<th>{options.get("label", "")}</th>'
        table_head += '</tr>'

        # Build table body
        table_body = ''
        for page in pages:
            data = page[1]
            path = data['_full_path']
            if not self._can_edit_page(path):
                continue
            table_body += '<tr>' \
                          f'<td class="path{" protected" if not self._can_edit_page(path) else ""}">' \
                          f'<a href="{self._editor_key}/edit/{path}" class="pure-menu-link">{path}</a>' \
                          '</td>'
            for key, options in config.get('columns', {}).items():
                value = data[key] if key in data else ""
                if 'options' in options:
                    for option in options['options']:
                        if 'values' in option and value in option['values']:
                            value = option['label']
                table_body += f'<td class="{key}"><span>{value}</span></td>'
            table_body += '</tr>'

        content = f'<div>{self._block_list_actions(config)}<h1>Pages</h1></div>' \
                  '<div class="page-list"><table class="pure-table pure-table-bordered">' \
                  f'<thead>{table_head}</thead>' \
                  f'<tbody>{table_body}</tbody>' \
                  '</table></div>'

        return self.to_html(content)

    def _block_list_actions(self, config: dict) -> str:
        """
        Adds the action buttons on the list page
        """
        item = ''
        for action in config.get('actions', {}):
            data = ''
            for key, value in action.get('data', {}).items():
                data += f' data-{key}="{self.__protect(value)}" '
            item += '<li class="pure-menu-item">' \
                    f'<a href="{self._editor_key}/{action.get("path", "")}" class="{action.get("class", "")}"{data}>' \
                    f'{action.get("label", "{label}")}' \
                    '</a>' \
                    '</li>'
        if item:
            return f'<div class="pure-menu pure-menu-horizontal actions"><ul class="pure-menu-list">{item}</ul></div>'
        return ''

    def page_edit(self, page_path: str = None) -> str:
        """
        Build the page form
        """
        if not self._can_edit_page(page_path):
            self.add_message('error', 'The page cannot be edited')

        config = self.get_config().get('form', {})

        # Init and load the page data
        if page_path is not None:
            data = self._fs.get_page_data(page_path)
            data['_page_type'] = self._fs.get_file_extension(page_path)
        else:
            data = self._fs.get_page_data()

        # Build the form
        columns = {"left": "", "right": ""}
        for fieldset in config.get('fieldsets', []):
            if not self.__can_show_fieldset(fieldset.get('depends', {}), data):
                continue
            columns[fieldset.get('column', 'left')] += self.__build_fieldset(fieldset, page_path, data)

        two_columns = columns.get('right') and columns.get('left')
        l_c = 'pure-u-1' + (' pure-u-md-7-24' if two_columns else '') + (' hide' if not columns.get('left') else '')
        r_c = 'pure-u-1' + (' pure-u-md-17-24' if two_columns else '') + (' hide' if not columns.get('right') else '')

        content = f'<form method="post" action="{self._editor_key}/edit/{page_path if page_path else ""}" ' \
                  'class="pure-form pure-form-stacked" enctype="multipart/form-data" id="edit_page">' \
                  '<h1>Page</h1>' \
                  f'{self._block_form_editor_function()}' \
                  f'{self._block_form_actions(page_path, config)}' \
                  f'<div class="{l_c} left"><div class="left-content">{columns.get("left", "")}</div></div>' \
                  f'<div class="{r_c} right"><div class="right-content">{columns.get("right", "")}</div></div>' \
                  '</form>' \
                  f'{self._form_ajax_send() if page_path else ""}'

        return self.to_html(content)

    def __build_fieldset(self, fieldset: dict, page_path: str, data: dict) -> str:
        """
        Build the fieldset
        """
        form = f'<fieldset><legend>{fieldset.get("label", "Fields")}</legend>'
        for var, options in fieldset.get('fields', {}).items():
            value = data[var] if var in data else ''
            if not self._can_edit_page(page_path) and not value:
                continue
            is_required = options.get('required', False)
            attributes = ' '.join([
                'disabled' if options.get('disabled', False) else '',
                'required' if is_required else '',
                'readonly' if options.get('readonly', False) else ''
            ]).strip()
            form += f'<div class="field pure-u-1{" hide" if options.get("hide", False) else ""}">' \
                    f'<label for="i_{var}" id="l_{var}">' \
                    f'{options.get("label", "Option")}{" <span>*</span>" if is_required else ""}' \
                    '</label>'
            f_type = options.get('type', 'text')
            if f_type in ['text']:
                form += f'<input type="text" id="i_{var}" name="{var}" value="{self.__protect(str(value))}"' \
                        f' class="pure-u-1 i-reload" {attributes}/>'
            if f_type in ['file', 'image']:
                if value:
                    form += f'<input type="checkbox" value="1" name="{var}_remove" class="i-reload" id="r_{var}" />' \
                            f'<label for="r_{var}" class="remove">Remove</label>'
                    if f_type == 'image':
                        form += f'<a href="/{value}" target="_blank" class="media">' \
                                f'<img src="/{value}" alt="{value}" height="70" />' \
                                f'</a>'
                        form += f'<figcaption class="pure-u-1">{value}</figcaption>'
                    else:
                        form += f'<a href="/{value}" target="_blank" class="media">{value}</a>'
                form += f'<input type="file" id="i_{var}" name="{var}" value=""' \
                        f' class="pure-u-1 i-reload" {attributes}/>'
            elif f_type in ['date']:
                form += f'<input type="date" id="i_{var}" name="{var}" value="{self.__protect(str(value))}"' \
                        f' class="pure-u-1 i-reload" {attributes}/>'
            elif f_type in ['textarea']:
                form += f'<textarea id="i_{var}" name="{var}"' \
                        f' class="pure-u-1 i-reload" {attributes}>{self.__protect(str(value))}</textarea>'
            elif f_type in ['select', 'multiselect']:
                multiple = 'multiple' if f_type == 'multiselect' else ""
                form += f'<select id="i_{var}" name="{var}" class="pure-u-1 i-reload" {multiple} {attributes}>'
                form += '<option value=""></option>'
                for option in options.get('options', []):
                    values = [str(value)] if not multiple else value
                    selected = 'selected="selected"' if str(option.get('value')) in values else ''
                    form += f'<option value="{self.__protect(str(option.get("value", "")))}" {selected}>' \
                            f'{str(option.get("label", ""))}' \
                            '</option>'
                form += '</select>'
            elif f_type in ['block']:
                file = None
                if value:
                    file = self._fs.get_source_dir(value)
                    if os.path.exists(file):
                        with open(os.path.normpath(file), 'r', encoding=self._fs.get_encoding()) as content:
                            value = content.read()
                            if self._fs.get_file_extension(file) == 'html':
                                code_elements = self.__extract_code(value)
                                for code in code_elements:
                                    value = value.replace(
                                        '<code' + code.get('attr') + '>' + code.get('content') + '</code>',
                                        '<code' + code.get('attr') + '>' + unescape(code.get('content')) + '</code>'
                                    )
                form += f'<textarea class="pure-u-1" id="i_{var}"' \
                        f' name="{var}" {attributes}>{escape(str(value))}</textarea>' \
                        '<span class="pure-form-message">' \
                        '<span>Ctrl-F / Cmd-F:</span> Start searching<br />' \
                        '<span>Ctrl-G / Cmd-G:</span> Find next<br />' \
                        '<span>Shift-Ctrl-G / Shift-Cmd-G:</span> Find previous<br />' \
                        '<span>Alt-G:</span> Jump to line<br /><br />' \
                        f'</span>' \
                        '<script type="text/javascript">' \
                        'initBlock(' \
                        f'"i_{var}",' \
                        f'"{options.get("mode", "html")}",' \
                        f'"{self._fs.get_file_extension(file) if file else ""}",' \
                        f'{"true" if options.get("focus", False) else "false"}' \
                        ')' \
                        '</script>'
            form += '</div>'
        form += '</fieldset>'
        return form

    @staticmethod
    def __extract_code(content: str) -> list:
        """
        Retrieve all code elements with attributes in the content
        """
        search = re.compile(r'<code(.*?)>(.*?)</code>', re.IGNORECASE | re.DOTALL).finditer(content)
        result = []
        for element in search:
            code = element.groups()
            result.append({'attr': code[0], 'content': code[1]})
        return result

    @staticmethod
    def __protect(value: str) -> str:
        """
        Protect the value by replacing special characters to HTML-safe sequences
        """
        return escape(unescape(value))

    @staticmethod
    def __can_show_fieldset(depends: dict, data: dict) -> bool:
        """
        Whether field set can be displayed based on field dependencies
        """
        can_show = True
        for depend in depends:
            if not depend.get('field') in data:
                can_show = False
            else:
                to_match = data[depend.get('field')]
                if not isinstance(to_match, list):
                    to_match = [to_match]
                if not any(item in to_match for item in depend.get('values')):
                    can_show = False
        return can_show

    @staticmethod
    def _form_ajax_send() -> str:
        """
        Save in Ajax when possible (only content updated)
        """
        return '<script type="text/javascript">' \
               'const form = document.getElementById("edit_page");' \
               'if (form) {' \
               'var ajax = true;' \
               'const fields = document.getElementsByClassName("i-reload");' \
               'for (var i = 0; i < fields.length; i++) {' \
               '  fields[i].addEventListener("change", function (event) {' \
               '    ajax = false;' \
               '  });' \
               '}' \
               'form.addEventListener("submit", function (event) {' \
               '  if (ajax) {' \
               '    event.preventDefault();' \
               '    fetch(form.action + "?ajax=1", {' \
               '      method: form.method,' \
               '      body: new FormData(form)' \
               '    }).then(function(response) {' \
               '      return response.text();' \
               '    }).then(function(html) {' \
               '      document.getElementById("system-messages").innerHTML = html;' \
               '      const messages = document.getElementsByClassName("system-message");' \
               '      setTimeout(function() {' \
               '        for (var i = 0; i < messages.length; i++) { messages[i].remove(); }' \
               '      }, 3000);' \
               '    });' \
               '  }' \
               '});' \
               '}' \
               '</script>'

    @staticmethod
    def _block_form_editor_function() -> str:
        """
        Init the editor for all block elements
        """
        return '<script type="text/javascript">' \
               'function initBlock(elementId, mode, extension, focus) {' \
               'const textarea = document.getElementById(elementId);' \
               'const width = textarea.offsetWidth;' \
               'const editor = CodeMirror.fromTextArea(' \
               '  textarea,' \
               '  {lineNumbers: true, lineWrapping: true, indentWithTabs: false, tabSize: 2}' \
               ');' \
               'textarea.nextSibling.style.width = width + "px";' \
               'if (focus) {' \
               '  editor.focus();' \
               '}' \
               'const modes = {html:"htmlmixed",md:"markdown",xml:"xml",css:"css",js:"javascript",json:"javascript"};' \
               'editor.setOption("mode", typeof modes[mode] !== "undefined" ? modes[mode] : null);' \
               'if (mode.charAt(0) === "$") {' \
               '  const types = document.getElementById(mode.replace("$", "i_"));' \
               '  if (!types && extension) {' \
               '    editor.setOption("mode", typeof modes[extension] !== "undefined" ? modes[extension] : null);' \
               '  }' \
               '  if (types) {' \
               '    const label = document.getElementById(elementId.replace("i_", "l_"));' \
               '    const labelText = label.innerText;' \
               '    if (extension) {' \
               '      types.value = extension;' \
               '    }' \
               '    label.innerText = types.options[types.selectedIndex].innerText || labelText;' \
               '    editor.setOption("mode", typeof modes[types.value] !== "undefined" ? modes[types.value] : null);' \
               '    types.addEventListener("change", function () {' \
               '      editor.setOption("mode", typeof modes[this.value] !== "undefined" ? modes[this.value] : null);' \
               '      label.innerText = this.options[this.selectedIndex].innerText || labelText;' \
               '    });' \
               '  }' \
               '}' \
               '}' \
               '</script>'

    def _block_form_actions(self, page_path: str or None, config: dict) -> str:
        """
        Menu with actions on the form page
        """
        actions = config.get('actions', {})
        menu = ''
        if actions.get('back', False):
            menu += f'<li class="pure-menu-item"><a href="{self._editor_key}/" class="pure-button">Back</a></li>'
        if actions.get('reset', False):
            menu += '<li class="pure-menu-item">' \
                    f'<a href="{self._editor_key}/edit/{page_path if page_path else ""}" class="pure-button">' \
                    'Reset' \
                    '</a>' \
                    '</li>'
        if page_path:
            if actions.get('data', False):
                menu += '<li class="pure-menu-item">' \
                        f'<a href="/_page/{page_path}" class="button-secondary pure-button" target="_blank">Data</a>' \
                        '</li>'
            if actions.get('view', False):
                menu += '<li class="pure-menu-item">' \
                        f'<a href="/{page_path}" class="button-secondary pure-button" target="_blank" data-cy="view">' \
                        f'View' \
                        f'</a>' \
                        '</li>'
        if page_path and self._can_edit_page(page_path) and actions.get('delete', False):
            menu += '<li class="pure-menu-item">' \
                    f'<a href="{self._editor_key}/remove/{page_path}" class="button-warning pure-button" ' \
                    'data-cy="delete" ' \
                    'onclick="javascript:return confirm(\'Are you sure you want to delete this page?\');">' \
                    'Delete' \
                    '</a>' \
                    '</li>'
        if self._can_edit_page(page_path) and actions.get('save', False):
            menu += '<li class="pure-menu-item">' \
                    '<input type="hidden" name="form_method" value="save" />' \
                    '<input type="submit" class="button-confirm pure-button" value="Save" data-cy="save" />' \
                    '</li>'
        if not menu:
            return ''
        return f'<div class="pure-menu pure-menu-horizontal actions"><ul class="pure-menu-list">{menu}</ul></div>'

    def _block_head_stylesheets(self) -> str:
        """
        Stylesheets
        """
        links = [self._editor_key + '/style.css']
        stylesheets = ''
        for css in links:
            stylesheets += f'<link rel="stylesheet" href="{css}" />'
        return stylesheets

    def _block_head_scripts(self) -> str:
        """
        Scripts
        """
        links = [self._editor_key + '/app.js']
        scripts = ''
        for script in links:
            scripts += f'<script type="text/javascript" src="{script}"></script>'
        return scripts

    @staticmethod
    def _block_head_style_inline() -> str:
        """
        CSS inline block
        """
        inline_style = [
            'body{font:normal 18px system-ui;padding:0;margin:0;background-color:#fcfcfc;background-image:url('
            '"data:image/svg+xml,%3Csvg xmlns=\'http://www.w3.org/2000/svg\' width=\'4\' height=\'4\' '
            'viewBox=\'0 0 4 4\'%3E%3Cpath fill=\'%23d7d7d7\' fill-opacity=\'0.4\' '
            'd=\'M1 3h1v1H1V3zm2-2h1v1H3V1z\'%3E%3C/path%3E%3C/svg%3E");}',
            'header{background:#000;font-size:.8em;width:100%;padding:6px 0 8px 0;color:#fff;line-height:26px}',
            'header .header-content{padding:0 4%}',
            'header a{color:#fff;text-decoration:none}',
            'header a:hover{color:#CCC}',
            'header ul{margin:0;padding:0}',
            'header li{display:inline-block;margin-right:20px}',
            'header .info{text-align:right}',
            'header .info span{margin:0 5px}',
            'header .switcher{text-align:center}',
            'header .switcher a{margin:0 5px;text-decoration:underline}',
            'header .switcher a.current{color:#3EAFD2}',
            'main{padding:10px 4% 100px}',
            'textarea{min-height:200px}',
            '.page-list{overflow-x:auto}',
            'table{width:100%;background:#fff;min-width:1200px;font-size:.9em}',
            'table .protected{background:#F4F4F4;color:#DADADA}',
            'table .protected a{color:#DADADA}',
            'label{font-weight:700;font-size:16px}',
            'label span{color:#A00}',
            'label.remove{display:inline;font-weight:400;margin-left:5px;cursor:pointer;font-size:15px}',
            '.right-content textarea{min-height:600px}',
            '.CodeMirror{border:1px solid #ccc;border-radius:4px;min-height:calc(100vh - 220px)}',
            '.CodeMirror-lines{line-height:28px}',
            '.CodeMirror-search-hint{display:none}',
            '.CodeMirror-dialog-top{padding:.5em .8em;font-size:16px;font-weight:700}',
            '.pure-form-stacked .CodeMirror-search-label{margin-top:0}',
            '.actions{position:fixed;right:0;left:0;bottom: 0;z-index:100;background:rgba(255,255,255,0.8);'
            ' padding:10px 4%;border-top:1px solid #ccc;text-align:right;white-space:normal}',
            '.actions .pure-menu-item{margin:5px 0 5px 20px}',
            '.pure-button{border-radius:5px}',
            '.button-warning{background:#C63B3B;color:#fff}',
            '.button-confirm{background:#1AAB3C;color:#fff}',
            '.button-secondary{background:#3EAFD2;color:#fff}',
            '.pure-table td{padding:0}',
            '.pure-table td span{display:inline-block;margin:.5em 1em}',
            '.pure-form-message{margin-top:.5em;font-size:.8em;line-height:24px}',
            '.pure-form-message span{color:#999}',
            '.system-message{position:fixed;top:0;left:0;right:0;z-index:100;padding:20px 4%;'
            ' border-bottom:1px solid #ccc;text-align:center;font-weight:bold;color:#fff}',
            '.system-message.success{background:#1AAB3C}',
            '.system-message.error{background:#C63B3B}',
            '.hide{display:none}',
            '.media{color:#333;display:block;margin:5px 0 10px 0;font-size:14px}',
            '.media img{max-width: 100%}',
            'figcaption{margin:0 0 10px 0;font-size:14px;font-style:italic}',
            '@media screen and (min-width: 48em){.left-content{padding-right:20px}table{font-size:1em}}',
            '@media screen and (max-width: 48em){.desktop{display:none}header .switcher{text-align:right}}'
        ]
        return "".join(inline_style)

    def _block_messages(self) -> str:
        """
        Messages block
        """
        html = ''
        success = self.get_messages('success')
        if success:
            html += '<div class="system-message success" data-cy="message-success">'
            for message in success:
                html += message
            html += '</div>'
        errors = self.get_messages('error')
        if errors:
            html += '<div class="system-message error" data-cy="message-error">'
            for message in errors:
                html += message
            html += '</div>'
        if errors or success:
            html += '<script type="text/javascript">' \
                    'const messages = document.getElementsByClassName("system-message");' \
                    'for(i=0;i<messages.length;i++){' \
                    'setTimeout(function(){messages[messages.length - 1].remove();}, 3000 + (i * 3000));' \
                    '}' \
                    '</script>'
        return html

    def to_html(self, content, custom_head: str = '') -> str:
        """
        The editor page template
        """
        return '<!DOCTYPE html><html>' \
               '<head><meta charset="utf-8" />' \
               '<meta name="viewport" content="width=device-width, initial-scale=1.0" />' \
               '<title>Editor</title>' \
               f'{self._block_head_scripts()}' \
               f'{self._block_head_stylesheets()}' \
               f'<style>{self._block_head_style_inline()}</style>{custom_head}</head>' \
               f'<body>' \
               f'{self._block_header()}' \
               f'<div id="system-messages">{self._block_messages()}</div>' \
               f'<main>{content}</main>' \
               '</body></html>'

    def _block_header(self):
        """
        Editor header HTML content
        """
        return '<header>' \
               '<div class="header-content pure-g">' \
               '<div class="pure-u-1-2 pure-u-md-1-3"><ul>' \
               f'<li><a href="{self._editor_key}">Pages</a></li>' \
               '<li><a href="/" target="_blank">Site</a></li>' \
               '<li><a href="https://www.stapy.net/editor.html" target="_blank" rel="noreferrer">Help</a></li>' \
               '</ul></div>' \
               f'<div class="pure-u-1-2 pure-u-md-1-3"><div class="switcher">{self._editor_switcher()}</div></div>' \
               '<div class="pure-u-1-3 desktop">' \
               '<div class="info">' \
               f'<span>Stapy {self._fs.get_version()}</span> &bull; <span>Editor {VERSION}</span>' \
               '</div>' \
               '</div>' \
               '</header>'

    def _editor_switcher(self):
        """
        Editor config switcher HTML content
        """
        switcher = self.get_config().get('editor_switcher', {})
        options = ''
        for label, key in switcher.items():
            link = self._path.replace(self._editor_key, key, 1)
            options += f'<a href="{link}" class="{"current" if self._editor_key == key else ""}">{label}</a>'
        return options

    def _can_edit_page(self, page_path: str or None) -> bool:
        """
        Test if the given page path can be edited
        """
        if not page_path:
            return True
        pages = self.get_config().get('protected_pages', [])
        for path in pages:
            if re.match(path, page_path):
                return False
        return True

    @staticmethod
    def _string_to_tuple(string: str) -> tuple or None:
        """
        Convert the string to tuple
        """
        return tuple(item for item in string.split(':')) if string else None

    @staticmethod
    def _format_path(path: str, replace_regex: str = '[^0-9a-z/]+') -> str:
        """
        Format the page path into a regular allowed path
        """
        path = path.lstrip('/')
        path = path.lower()
        nfkd_form = unicodedata.normalize('NFKD', path)
        path = "".join([c for c in nfkd_form if not unicodedata.combining(c)])
        path = re.sub(replace_regex, '-', path)
        path = path.strip('-')
        return path
